
from .useful import recoverage

class AsueRequestBody:
    
    def __init__(self, body: dict):
        if isinstance(body, dict):
            for k in body.items():
                key = recoverage(k[0])
                self.__setattr__(key, k[1])
                
                
    def get_json(self):
        return self.__dict__




from .models.routes import AsureBaseClassRoute
from .models.websocket import AsureBaseWebsocket
from .template import AsueTemplateRender
from .request import AsueRequestObject
from .resource import AsueRoute
from .primitive import Handler

import nest_asyncio
import uvicorn 
import asyncio
import json

nest_asyncio.apply()

h = Handler()

async def server(scope, receive, send):
    if scope['type'] == 'http':
        request = AsueRequestObject(scope=scope, receive=receive, send=send)
        for resource in Asure.get_routes():       
            await h.inject_routes_dependencies(request=request, resource=resource)  
    if scope['type'] == 'websocket' and Asure.has_ws():
        await h.get_ws_connection(scope=scope, receive=receive, send=send, ws_handler=Asure.get_ws())
   
class Asure:
    
    RESOURCES = []
    ws = None
    currently_app = None
    
    def __init__(self, templates_dir :str=None, statics_dir :str=None):
                
        Asure.set_currently_app(app=self)          
        self.atr = AsueTemplateRender()       
        if templates_dir:    
            self.atr.get_directory_templates(templates_dir)
        if statics_dir:
            self.atr.get_directory_statics(statics_dir)
            
    
    @classmethod    
    def set_currently_app(cls, app):
        if cls.currently_app == None:
            cls.currently_app = app
       
        
    @classmethod
    def get_currently_app(cls):
        return cls.currently_app 
        
        
    @classmethod
    def set_ws(cls, ws):
        cls.ws = ws
        
        
    @classmethod
    def get_ws(cls):
        return cls.ws
        
        
    @classmethod
    def has_ws(cls):
        if cls.ws!=None:
            return True 
        else:
            return False
        
        
    @classmethod
    def get_routes(cls):
        return cls.RESOURCES 
     
     
    def add_resource(self, 
        route_class=None,
        route_path :str=None,
        websocket_resource :AsureBaseWebsocket=None):
        
        if route_class and route_path: 
            AsueRoute.register(instance=route_class(route=route_path), resources=Asure.get_routes())
        if websocket_resource: 
            Asure.set_ws(ws=websocket_resource())
                                              
                           
    def run(self, host: str, port: int):  
        uvicorn.run("asure:server", log_level="info", host=host, port=port)


from ..template.useful import encode_data, decode_data
from ..template import AsueTemplateRender

import json

class AsueRequestResponse:
    def __init__(self, send :callable):
        self._send = send
        
        
    async def send(self, response_json :dict=None, response_html :bytes=None):
        
        if response_json:
            await self._send(self._get_first_json_package(response_json=response_json))
            await self._send(self._get_second_json_package(response_json=response_json)) 
             
        if response_html:
            await self._send(self._get_first_html_package())
            await self._send(self._get_second_html_package(response_html=response_html))
            
        
        
    async def renderize_html_page(self, template_name :str):
        self.str = AsueTemplateRender.get_instance()
        html_page_encoded = self.str.render_template(template_name=template_name)
        await self.send(response_html=html_page_encoded)
        
        
    def _get_first_html_package(self):
        first_package = {}
        first_package.update({'type': 'http.response.start'})
        first_package.update({'status': 200})  
        first_package.update({'headers': [
                [encode_data('content-type'), encode_data('text/html')]
            ]
        })  
        return first_package
    
    
    def _get_second_html_package(self, response_html :bytes):
        second_package = {}  
        second_package.update({'type': 'http.response.body'})  
        second_package.update({'body': response_html})
        return second_package
        
        
    def _get_first_json_package(self, response_json :dict):
        first_package = {}  
        first_package.update({'type': 'http.response.start'})
        if 'status' in response_json:
            first_package.update({'status': response_json['status']})
        else:
            first_package.update({'status': 200})
            
        if 'headers' in response_json:
            headers = [] 
            if 'content-type' in response_json['headers']:
                headers.append([
                    encode_data('content-type'), 
                    encode_data(response_json['headers']['content-type'])
                    ])
            first_package.update({'headers': headers})
        else:
            first_package.update({'headers': [
                [encode_data('content-type'), encode_data('application/json')]
                ]
            })          
        return first_package
    
    
    def _get_second_json_package(self, response_json :dict):
        second_package = {}  
        second_package.update({'type': 'http.response.body'})
        if 'status' in response_json:
            del response_json['status']
        if 'headers' in response_json:
            del response_json['headers']
        second_package.update({'body': encode_data(json.dumps(response_json))})
        return second_package
        
        
        


from ..template.useful import decode_data, encode_data
from .useful import recoverage

class AsueRequestHeaders:
    
    def __init__(self, headers: dict):
        if isinstance(headers, dict):
            for k in headers.items(): 
                key = recoverage(k[0])
                self.__setattr__(key, k[1])
                                       
    def get_json(self):
        return self.__dict__


from .useful import encode_data, decode_data
from string import Template
import os

class AsueTemplateRender:
    
    instance = None
    totaly_files_dir = {}
    
    def __init__(self):
        AsueTemplateRender.set_instance(instance=self)
    
    
    @classmethod
    def set_instance(cls, instance):
        if cls.instance == None:
            cls.instance = instance
        
    
    @classmethod
    def get_instance(cls):
        if cls.instance != None:
            return cls.instance
    
        
    def get_file_dir(self, line :str):
        file_link = line.split('*')[1]
        file_dir = file_link.split('"')[-2]
        file_name = file_dir.split('/')[-1]
        totaly_files_dir = AsueTemplateRender.get_totaly_files_dir()
        for item in totaly_files_dir.items():
            for _dir in item[1]:
                if file_name in _dir:
                    return _dir
    
    def open_file(self, file_dir :str):
        with open(file_dir, 'r') as static_file:
            static_file_list = []
            for line in static_file.readlines():
                static_file_list.append(line)
            return static_file_list

    def check_break_line(self, line :str):
        if '\n' in line:
            return line 
        else:
            return line + '\n'
    
    def replace_style(self, line :str, file_list :list):       
        style_init_structure = '<style type="text/css">\n'
        style_end_structure = '</style>\n'       
        style_structure = []    
        style_structure.append(style_init_structure)
        for lin in file_list:
            style_structure.append(self.check_break_line(line=lin))        
        style_structure.append(style_end_structure)
        style = ''
        for i in style_structure:
            style = style + i          
        return style
    
    def replace_script(self, line :str, file_list :list):
        script_init_structure = '<script type="text/javascript">\n'
        script_end_structure = '</script>\n'  
        script_structure = []
        script_structure.append(script_init_structure)
        for lin in file_list:
            script_structure.append(self.check_break_line(line=lin))
        script_structure.append(script_end_structure)
        script = ''
        for i in script_structure:
            script = script + i 
        return script
    
    
    def insert_static_files(self, line :str, template :list):     
        if 'style(*' in line and '*)' in line:
            css_file_dir =  self.get_file_dir(line=line)
            css_file_list = self.open_file(css_file_dir)
            css_structure = self.replace_style(line=line, file_list=css_file_list)
            return css_structure    
        if 'script(*' in line and '*)' in line:
            js_file_dir = self.get_file_dir(line=line)
            js_file_list = self.open_file(js_file_dir)
            js_structure = self.replace_script(line=line, file_list=js_file_list)
            return js_structure        
        return line
    
    
    @classmethod
    def set_file_dir(cls, key :str, value :list):
        cls.totaly_files_dir.update({key: value})
        
    
    @classmethod
    def get_totaly_files_dir(cls):
        return cls.totaly_files_dir
    
    
    def get_directory_templates(self, template_dir :str):
        html_files_dir = []
        for root, subfolder, filename in os.walk(template_dir):
            for filen in filename:
                if '.html' in filen:
                    html_files_dir.append(os.path.join(root, filen)) 
                    
        AsueTemplateRender.set_file_dir(key='html_files_dir', value=html_files_dir)
    
    def get_css_files_dir(self, static_dir :str):
        css_files_dir = []
        for root, subfolder, filename in os.walk(static_dir):
            for filen in filename:
                if '.css' in filen:
                    css_files_dir.append(os.path.join(root, filen))
        
        AsueTemplateRender.set_file_dir(key='css_files_dir', value=css_files_dir)
                      
                    
    def get_js_files_dir(self, static_dir :str):
        js_files_dir = []
        for root, subfolder, filename in os.walk(static_dir):
            for filen in filename:
                if '.js' in filen:
                    js_files_dir.append(os.path.join(root, filen)) 
                    
        AsueTemplateRender.set_file_dir(key='js_files_dir', value=js_files_dir)    
    
    
    def get_directory_statics(self, static_dir :str):
        self.get_css_files_dir(static_dir=static_dir)  
        self.get_js_files_dir(static_dir=static_dir)
        
        
    def _read_template(self, template_name: str):
        totaly_files_dir = AsueTemplateRender.get_totaly_files_dir()
        for html_template_dir in totaly_files_dir['html_files_dir']:
            if template_name in html_template_dir:          
                with open('./'+html_template_dir, 'r') as template:             
                    nTemplate = [] 
                    for line in template.readlines():
                        nTemplate.append(line)                      
                    page = ''
                    for line in nTemplate:                  
                        page = page + self.insert_static_files(line=line, template=nTemplate) 
                                      
                    return encode_data(page) 
        
        
    def _read_and_insert_datas_into_template(self, template_path :str, context_datas :dict=None):
        if not isinstance(context_datas, dict):
            print('Valor setado em "context_datas" não é do tipo "dict"')
        if not isinstance(template_path, str):
            print('Valor setado em "template_path" não é do tipo "str"')     
        try:        
            with open(f"{self.templates_directory}{template_path}", 'r') as template:              
                new_template = [] 
                for line in template.readlines():
                    new_template.append(line)               
                ste = SpartacusTemplateEngine()
                ste.insert(template_block_total=new_template, context_datas=context_datas)               
                page = ''
                for line in new_template:
                    page = page + line
                return encode_data(page) 
        except FileNotFoundError:
            raise 'Template Not found'


    def render_template(self, template_name, context_datas :dict=None):     
        if context_datas!=None:                
            return self._read_and_insert_datas_into_template(
                context_datas=context_datas, 
                template_path=template
            )                 
        else:                     
            return self._read_template(
                template_name=template_name
            )



import json


class AsueRoute:
    
    def __init__(
        self, 
        path :str,
        instance :object,
        ):
        self.path = path
        self.instance = instance
        
    @classmethod   
    def register_resource(
        cls,
        resources :list, 
        instance :object
        ): 
        resources.append(AsueRoute(
            instance=instance, 
            path=instance.route
        ))    
        
    @classmethod 
    def register(cls, instance :object, resources :list):
        
        cls.register_resource(resources=resources, instance=instance)    
            
    def get_json_route(self):
        return self.__dict__
    
class AsueWebsocket:
    
    def __init__(self, send, receive, scope):
        self._scope = scope
        self._send = send 
        self._receive = receive
        
    async def recv(self):
        order = await self._receive()
        if order['type']=='websocket.connect':
            await self._send({'type': 'websocket.accept'})
            order = await self._receive()
            if order['type']=='websocket.receive':
                return json.loads(order['text'])
            elif order['type']=='websocket.disconnect': 
                await self._send({'type': 'websocket.close'})
    
    async def emit(self, data):
        await self._send({'type': 'websocket.send', 'text': data})





from .body import AsueRequestBody
from .params import AsueRequestParams
from .headers import AsueRequestHeaders 
from .response import AsueRequestResponse
from ..template.useful import decode_data, encode_data
import asyncio
import json
from json.decoder import JSONDecodeError
            

class AsueRequestObject:
    
    def __init__(self, scope :dict, receive :callable, send :callable):
           
        self.path = self.get_path(scope=scope)
        self.method = self.get_method(scope=scope) 
        self.responser = AsueRequestResponse(send=send)  
        self.coroutine(function=self.read_body(receive=receive)) 
        self.params = AsueRequestParams(params=self.get_params(scope=scope))      
        self.headers = AsueRequestHeaders(headers=self.get_headers(scope=scope))         
              
           
    def get_responser(self, send):
        return send
    
    def get_path(self, scope):
        return scope['path']
    
    def get_method(self, scope):
        return scope['method']
        
    def get_headers(self, scope):       
        headers = scope['headers']    
        decodeds = []
        for item in headers:       
            item = (decode_data(item[0]), decode_data(item[1]))        
            decodeds.append(item)        
        newl = {}
        for item in decodeds:        
            newl.update({item[0] : item[1]})                                      
        return newl
    
    
    def get_params(self, scope):       
        params = decode_data(scope['query_string'])    
        params = params.split('&')       
        aux = []    
        for item in params:        
            aux.append(item.split('='))        
        newl = {}    
        for item in aux:        
            if len(item)>1:            
                newl.update({item[0] : item[1]})                 
        return newl
        
        
    async def read_body(self, receive):        
        body = b''        
        more_body = True       
        while more_body:        
            message = await receive()         
            body += message.get('body', b'')          
            more_body = message.get('more_body', False)
        nBody = ''
        try:                   
            nBody = json.loads(body)   
        except JSONDecodeError:
            pass
        self.body = AsueRequestBody(body=nBody) 
          
    
    def coroutine(self, function):
        loop = asyncio.new_event_loop()     
        asyncio.set_event_loop(loop)        
        loop.run_until_complete(function) 
    

    def get_json(self):
        newd = self.__dict__
        headers = newd['headers']
        params = newd['params']
        body = newd['body']
        newd['headers'] = headers.get_json()
        newd['params'] = params.get_json()
        newd['body'] = body.get_json()    
        newdd = {}
        for item in newd.items():
            newdd.update({item[0]: item[1]})         
        del newdd['responser']
        return json.dumps(newdd, indent=2)


# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['asure',
 'asure.exceptions',
 'asure.models',
 'asure.models.routes',
 'asure.models.websocket',
 'asure.primitive',
 'asure.request',
 'asure.resource',
 'asure.template']

package_data = \
{'': ['*']}

install_requires = \
['uvicorn==0.12.2']

setup_kwargs = {
    'name': 'asure',
    'version': '0.1.0',
    'description': '',
    'long_description': None,
    'author': 'JorgeFernando',
    'author_email': 'jorgebg2016@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from abc import abstractmethod
from ...request import AsueRequestObject
from ...request.response import AsueRequestResponse


class AsureBaseClassRoute:
  
    def __init__(self, route :str):
        self.route = route
        
    
    @abstractmethod
    async def view(self, response :AsueRequestResponse, request :AsueRequestObject):
        pass
    
    @abstractmethod
    async def get(self, response :AsueRequestResponse, request :AsueRequestObject):
        pass 
    
    @abstractmethod
    async def post(self, response :AsueRequestResponse, request :AsueRequestObject):
        pass
    
    @abstractmethod
    async def put(self, response :AsueRequestResponse, request :AsueRequestObject):
        pass 
    
    @abstractmethod
    async def delete(self, response :AsueRequestResponse, request :AsueRequestObject):
        pass
    
    @abstractmethod
    async def patch(self, response :AsueRequestResponse, request :AsueRequestObject):
        pass
    
    @abstractmethod
    async def head(self, response :AsueRequestResponse, request :AsueRequestObject):
        pass
    
    @abstractmethod
    async def trace(self, response :AsueRequestResponse, request :AsueRequestObject):
        pass
    
    @abstractmethod
    async def connect(self, response :AsueRequestResponse, request :AsueRequestObject):
        pass


from ..resource import AsueWebsocket, AsueRoute
from .useful import encode_data, decode_data
from ..request import AsueRequestObject
import asyncio
import json

class Handler:

    async def inject_routes_dependencies(self, request: AsueRequestObject, resource: AsueRoute):    
        
        if request.path==resource.path:      
                     
            if request.method=='VIEW':
                try:                  
                    await resource.instance.view(response=request.responser, request=request)   
                except:
                    raise 'Route is not defined'                
            if request.method=='POST':                   
                await resource.instance.post(response=request.responser, request=request)                  
            if request.method=='GET':                
                await resource.instance.get(response=request.responser, request=request)                
            if request.method=='DELETE':               
                await resource.instance.delete(response=request.responser, request=request)                
            if request.method=='PUT':               
                await resource.instance.put(response=request.responser, request=request)                 
            if request.method=='PATCH':               
                await resource.instance.patch(response=request.responser, request=request)              
                                  
    
    async def get_ws_connection(self, scope, receive, send, ws_handler):
              
        await ws_handler.handler(
            websocket= AsueWebsocket(
                receive=receive,
                scope=scope, 
                send=send
            )
        )


    def to_response(self, send, 
        content_type :str='application/json',                   
        json_response :dict=None, template :bytes=None):
    
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)     
        loop.run_until_complete(self._response(         
            send=send,        
            template=template,       
            json_response=json_response,         
            content_type=content_type         
        ))
        

    async def _response(self, send, content_type :str, template :bytes=None, json_response :dict=None):   
        
        if template:        
            response_data = template        
        elif json_response:      
            response_data = json_response           
        if template==None and json_response==None:   
            raise 'Response no defined'      
         
        if isinstance(response_data, dict):      
            response_data = encode_data(json.dumps(response_data))          
        try:        
            self.send_in_order(function=self.send_first_package(      
                response_data=response_data,
                content_type=content_type,
                send=send
                )
            )                  
        except RuntimeError:      
            pass